/*
 * Copyright (C) 2020, IOPSYS Software Solutions AB.
 *
 * Author: Vivek Dutta <vivek.dutta@iopsys.eu>
 * Co-Author: Shubham Sharma <shubham.sharma@iopsys.eu>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include "vendor_uspd.h"
#include "os_utils.h"
#include "common_defs.h"
#include "data_model.h"
#include "str_vector.h"
#include "json.h"
#include "text_utils.h"

#include <libubus.h>
#include <libubox/blobmsg.h>
#include <libubox/blobmsg_json.h>

// Timeout in milliseconds
#define USP_PROTO "usp"
#define USPD_TIMEOUT 10000
#define MIN_NUM_TO_GROUP (10)
#define NUMENTRY "NumberOfEntries"
#define MAX_CACHE_TIME (30)
#define MAX_GROUP_SEP 2
#define DM_VENDOR_MOBJ        0x01000000
#define ROOT_PATH "Device."

extern int CountPathSeparator(char *path);
static struct ubus_context *g_ubus_ctx;
static str_vector_t g_group_vec;
static int uspd_mq_sockets[2] = {-1, -1};
static kv_vector_t g_cached_kv;
static int g_cached_index;
static int g_curr_trans_id = 0;

#define mq_rx_socket  uspd_mq_sockets[0]
#define mq_tx_socket  uspd_mq_sockets[1]
#define IS_UBUS_Q_READY ((mq_tx_socket == -1) ? false : true)
#define STR(s) #s

pthread_cond_t g_wait_ubus = PTHREAD_COND_INITIALIZER;
pthread_mutex_t g_ubus_lock = PTHREAD_MUTEX_INITIALIZER;

typedef enum
{
	CMD_NULL,
	CMD_GROUP_GET,
	CMD_GROUP_SET,
	CMD_ADD,
	CMD_DEL,
	CMD_GET,
	CMD_SET,
	CMD_OPERATE_SYNC,
	CMD_OPERATE_ASYNC,
	CMD_INSTANCES,
	CMD_TRAN_START,
	CMD_TRAN_COMMIT,
	CMD_TRAN_ABORT,
} uspd_cmd_t;

typedef struct
{
	uspd_cmd_t cmd;
	int len;
	int inst;
	int fault;
	char *path;
	char *p_value;
	str_vector_t *vec;
	kv_vector_t *kv;
	kv_vector_t *kv_out;
	unsigned *p_types;
	int *p_instance;
	int *p_failure_index;
} vendor_data_t;

// Private functions
char *stringify_cmd(int cmd)
{
	switch(cmd) {
	case CMD_NULL:
		return STR(CMD_NULL);
	case CMD_GROUP_GET:
		return STR(CMD_GROUP_GET);
	case CMD_GROUP_SET:
		return STR(CMD_GROUP_SET);
	case CMD_ADD:
		return STR(CMD_ADD);
	case CMD_DEL:
		return STR(CMD_DEL);
	case CMD_GET:
		return STR(CMD_GET);
	case CMD_SET:
		return STR(CMD_SET);
	case CMD_OPERATE_SYNC:
		return STR(CMD_OPERATE_SYNC);
	case CMD_OPERATE_ASYNC:
		return STR(CMD_OPERATE_ASYNC);
	case CMD_INSTANCES:
		return STR(CMD_INSTANCES);
	case CMD_TRAN_START:
		return STR(CMD_TRAN_START);
	case CMD_TRAN_COMMIT:
		return STR(CMD_TRAN_COMMIT);
	case CMD_TRAN_ABORT:
		return STR(CMD_TRAN_ABORT);
	}

	return NULL;
}
static int uspd_call(struct ubus_context *ctx, char *method,
		     struct blob_buf *data, ubus_data_handler_t callback,
		     vendor_data_t *arg)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;

	fault = ubus_lookup_id(ctx, "usp.raw", &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] usp.raw not present (%s)",
			      __func__, __LINE__, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	// Invoke Ubus to get data from uspd
	fault = ubus_invoke(ctx, id, method, data->head, callback, arg, USPD_TIMEOUT);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus call for (%s)method failed(%s)",
			      __func__, __LINE__, method, ubus_strerror(fault));
		if (arg)
			arg->fault = USP_ERR_INTERNAL_ERROR;
		return USP_ERR_INTERNAL_ERROR;
	}

	return USP_ERR_OK;
}

static int uspd_call_inline_nocb(char *method, struct blob_buf *bb)
{
	struct ubus_context *ctx = ubus_connect(NULL);
	if (!ctx) {
		USP_LOG_Error("[%s:%d] ubus_connect failed", __func__, __LINE__);
		return USP_ERR_INTERNAL_ERROR;
	}

	uspd_call(ctx, method, bb, NULL, NULL);

	ubus_free(ctx);
	return USP_ERR_OK;
}

static void uspd_ubus_complete_cb(struct ubus_request *req, int ret)
{
    USP_FREE(req);
}

static int uspd_call_async(char *method, struct blob_buf *data, ubus_data_handler_t callback, void *priv)
{
	uint32_t id;
	int fault = UBUS_STATUS_OK;
	struct ubus_request *req;

	fault = ubus_lookup_id(g_ubus_ctx, "usp.raw", &id);
	if (fault) {
		USP_LOG_Error("[%s:%d] usp.raw not present (%s)",
			      __func__, __LINE__, ubus_strerror(fault));
		return USP_ERR_INTERNAL_ERROR;
	}

	req = USP_MALLOC(sizeof(struct ubus_request));
	if (req == NULL) {
		USP_LOG_Error("Out of memory!");
		return USP_ERR_RESOURCES_EXCEEDED;
	}

	memset(req, 0, sizeof(struct ubus_request));
	fault = ubus_invoke_async(g_ubus_ctx, id, method, data->head, req);
	if (fault) {
		USP_LOG_Error("[%s:%d] ubus async call failed(%s)",
			      __func__, __LINE__, ubus_strerror(fault));
		return USP_ERR_INTERNAL_ERROR;
	}

	req->data_cb = callback;
	req->complete_cb = uspd_ubus_complete_cb;
	req->priv = priv;

	ubus_complete_request_async(g_ubus_ctx, req);

	return USP_ERR_OK;
}

static struct blob_attr *get_parameters(struct blob_attr *msg)
{
	struct blob_attr *params = NULL;
	struct blob_attr *cur;
	int rem;

	blobmsg_for_each_attr(cur, msg, rem) {
		if (blobmsg_type(cur) == BLOBMSG_TYPE_ARRAY) {
			params = cur;
			break;
		}
	}
	return params;
}

static int detect_fault(struct blob_attr *msg)
{
	int fault = USP_ERR_OK;
	enum {
		FAULT_ID,
		__FAULT_MAX
	};
	struct blob_attr *f_attr[__FAULT_MAX] = {NULL};
	const struct blobmsg_policy fp[__FAULT_MAX] = {
		{ "fault", BLOBMSG_TYPE_INT32 }
	};

	blobmsg_parse(fp, __FAULT_MAX, f_attr, blobmsg_data(msg),
			    blobmsg_len(msg));
	if (f_attr[FAULT_ID])
		fault = blobmsg_get_u32(f_attr[FAULT_ID]);

	return fault;
}

static void get_parameters_in_kv(struct blob_attr *params, kv_vector_t *kv_out)
{
	struct blob_attr *cur;
	size_t rem;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING },
	};

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_PARAM] && tb[P_VALUE]) {
			USP_ARG_Add(kv_out,
				    blobmsg_get_string(tb[P_PARAM]),
				    blobmsg_get_string(tb[P_VALUE]));
		}
	}

}

static void _sync_operate_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *params, *results, *rtemp;
	kv_vector_t *kv_out;
	size_t rem;

	if (!msg)
		return;

	kv_out = arg->kv_out;

	results = get_parameters(msg);
	if (results == NULL)
		return;

	blobmsg_for_each_attr(rtemp, results, rem) {
		arg->fault = detect_fault(rtemp);
		if (arg->fault != USP_ERR_OK)
			return;

		params = get_parameters(rtemp);
		if (params == NULL)
			return;

		get_parameters_in_kv(params, kv_out);
		break;
	}
}

static void _add_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	size_t rem;
	struct blob_attr *results, *cur;
	vendor_data_t *arg = req->priv;
	enum {
		P_INSTANCE,
		P_FAULT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "instance", BLOBMSG_TYPE_STRING },
		{ "fault", BLOBMSG_TYPE_INT32 },
	};

	if (!msg)
		return;

	results = get_parameters(msg);
	if (results == NULL)
		return;

	blobmsg_for_each_attr(cur, results, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT])
			arg->fault = blobmsg_get_u32(tb[P_FAULT]);

		if (tb[P_INSTANCE])
			*arg->p_instance = atoi(blobmsg_get_string(tb[P_INSTANCE]));
		break;
	}

}

static void _set_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;
	enum {
		P_FAULT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "fault", BLOBMSG_TYPE_INT32 }
	};

	if (!msg)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT]) {
			arg->fault = blobmsg_get_u32(tb[P_FAULT]);
			break;
		}
	}
}

static void _set_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	vendor_data_t *arg = req->priv;
	enum {
		P_PATH,
		P_FAULT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "path", BLOBMSG_TYPE_STRING },
		{ "fault", BLOBMSG_TYPE_INT32 }
	};

	if (!msg)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_FAULT] && tb[P_PATH]) {
			char *path = blobmsg_get_string(tb[P_PATH]);

			arg->fault = blobmsg_get_u32(tb[P_FAULT]);
			*arg->p_failure_index = KV_VECTOR_FindKey(arg->kv, path, 0);
			break;
		}
	}
}

static void _del_object_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	_set_value_single_cb(req, type, msg);
}

static int _set_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "value", arg->p_value);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	uspd_call(g_ubus_ctx, "set", &bb, _set_value_single_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static int _set_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	void *array, *table;
	int i;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	array = blobmsg_open_array(&bb, "pv_tuple");
	for (i = 0; i < arg->kv->num_entries; i++) {
		table = blobmsg_open_table(&bb, NULL);
		blobmsg_add_string(&bb, "path", arg->kv->vector[i].key);
		blobmsg_add_string(&bb, "value", arg->kv->vector[i].value);
		blobmsg_close_table(&bb, table);
	}
	blobmsg_close_array(&bb, array);

	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	uspd_call(g_ubus_ctx, "setm_values", &bb, _set_value_group_cb, arg);

	blob_buf_free(&bb);
	return USP_ERR_OK;
}

static int _add_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	uspd_call(g_ubus_ctx, "add_object", &bb, _add_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static int _del_object(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	uspd_call(g_ubus_ctx, "del_object", &bb, _del_object_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static int _operate_sync_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	char path[MAX_DM_PATH] = {'\0'}, *action;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	if (arg->kv_out)
		KV_VECTOR_Init(arg->kv_out);

	// separate path and command
	action = TEXT_UTILS_SplitPath(arg->path, path, MAX_DM_PATH);
	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv_pair_t *kv;

			kv = &arg->kv->vector[i];
			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	blobmsg_add_string(&bb, "path", path);
	blobmsg_add_string(&bb, "action", action);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	uspd_call(g_ubus_ctx, "operate", &bb, _sync_operate_cb, arg);

	blob_buf_free(&bb);

	return arg->fault;
}

static void _operate_async_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params, *results, *rtemp;
	int *instance = (int *) req->priv;
	char err_log[128] = "Error not specified by vendor";
	char *err_msg;
	int fault = USP_ERR_OK;
	kv_vector_t *output_args;
	size_t rem;

	if (!msg)
		return;

	results = get_parameters(msg);
	if (results == NULL)
		return; // TODO: handler error case

	output_args = USP_ARG_Create();
	blobmsg_for_each_attr(rtemp, results, rem) {
		fault = detect_fault(rtemp);
		if (fault != USP_ERR_OK)
			break;

		params = get_parameters(rtemp);
		if (params == NULL)
			break;

		get_parameters_in_kv(params, output_args);
		break;
	}


	USP_LOG_Info("## Dump async out arg:");
	KV_VECTOR_Dump(output_args);

	err_msg = (fault != USP_ERR_OK) ? err_log : NULL;
	USP_SIGNAL_OperationComplete(*instance, fault, err_msg, output_args);
	USP_SAFE_FREE(instance);
}

static int _operate_async_handler(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	int fault = USP_ERR_OK;
	kv_pair_t *kv;
	char path[MAX_DM_PATH] = {'\0'}, *action;
	int *instance;

	instance = (int *) USP_MALLOC(sizeof(int));
	if (instance == NULL)
		return USP_ERR_COMMAND_FAILURE;

	// Exit if unable to signal that this operation is active
	fault = USP_SIGNAL_OperationStatus(arg->inst, "Active");
	if (fault != USP_ERR_OK) {
		return USP_ERR_COMMAND_FAILURE;
	}

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// separate path and command
	action = TEXT_UTILS_SplitPath(arg->path, path, MAX_DM_PATH);

	blobmsg_add_string(&bb, "path", path);
	blobmsg_add_string(&bb, "action", action);
	blobmsg_add_string(&bb, "proto", USP_PROTO);
	blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);

	if (arg->kv != NULL) {
		void *table = blobmsg_open_table(&bb, "input");

		for (int i = 0; i < arg->kv->num_entries; ++i) {
			kv = &arg->kv->vector[i];
			blobmsg_add_string(&bb, kv->key, kv->value);
		}
		blobmsg_close_table(&bb, table);
	}

	*instance = arg->inst;
	arg->fault = uspd_call_async("operate", &bb, _operate_async_cb, instance);
	//TODO: handle error case here as well

	blob_buf_free(&bb);
	return fault;
}

static void _tran_exec_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	struct blob_attr *f_attr[1] = {NULL};
	const struct blobmsg_policy fp[1] = {
		{ "status", BLOBMSG_TYPE_BOOL }
	};

	g_curr_trans_id = 0;

	if (!msg)
		return;

	blobmsg_parse(fp, 1, f_attr, blobmsg_data(msg), blobmsg_len(msg));
	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (status == false) {
		USP_LOG_Info("Another transaction in progress");
		arg->fault = USP_ERR_INTERNAL_ERROR;
	}
}

static void _tran_start_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	bool status = false;
	int trans_id = 0;
	struct blob_attr *f_attr[2] = {NULL};
	const struct blobmsg_policy fp[2] = {
		{ "status", BLOBMSG_TYPE_BOOL },
		{ "transaction_id", BLOBMSG_TYPE_INT32 }
	};

	if (!msg)
		return;

	blobmsg_parse(fp, 2, f_attr, blobmsg_data(msg), blobmsg_len(msg));

	if (f_attr[0])
		status = blobmsg_get_u8(f_attr[0]);

	if (f_attr[1])
		trans_id = blobmsg_get_u32(f_attr[1]);

	if (status == false) {
		USP_LOG_Info("Another transaction in progress");
		arg->fault = USP_ERR_INTERNAL_ERROR;
	} else {
		g_curr_trans_id = trans_id;
	}
}

static int _tran_exec(vendor_data_t *arg, char *cmd)
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	// Invoke Ubus to get data from uspd
	if (strcmp(cmd, "transaction_start") == 0) {
		blobmsg_add_string(&bb, "app", "obuspa");
		uspd_call(g_ubus_ctx, cmd, &bb, _tran_start_cb, arg);
	} else {
		blobmsg_add_u32(&bb, "transaction_id", g_curr_trans_id);
		uspd_call(g_ubus_ctx, cmd, &bb, _tran_exec_cb, arg);
	}

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_value_single_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING }
	};

	if (!msg)
		return;

	arg->fault = detect_fault(msg);
	if (arg->fault)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};
		char *path, *val;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_VALUE])
			continue;

		path = blobmsg_get_string(tb[P_PARAM]);
		val = blobmsg_get_string(tb[P_VALUE]);

		if (strcmp(path, arg->path) == 0) {
			USP_STRNCPY(arg->p_value, val, arg->len);
			break;
		}
	}
}

int _get_value_single(vendor_data_t *arg)
{
	struct blob_buf bb = { };

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", "usp");

	// Invoke Ubus to get data from uspd
	uspd_call(g_ubus_ctx, "get", &bb, _get_value_single_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

static void _get_value_group_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	int index, i;
	vendor_data_t *arg = req->priv;
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;
	kv_vector_t pv_set;
	kv_pair_t *pair;
	enum {
		P_PARAM,
		P_VALUE,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "value", BLOBMSG_TYPE_STRING }
	};

	if (!msg)
		return;

	arg->fault = detect_fault(msg);
	if (arg->fault)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	index = 0;
	USP_ARG_Init(&pv_set);
	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL, NULL};
		char *path, *val;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_VALUE])
			continue;

		path = blobmsg_get_string(tb[P_PARAM]);
		val = blobmsg_get_string(tb[P_VALUE]);

		USP_ARG_Add(&pv_set, path, val);
	}

	for (i = 0, index = 0; i < arg->kv->num_entries; i++) {
		pair = &arg->kv->vector[i];
		index = KV_VECTOR_FindKey(&pv_set, pair->key, index);
		if (index != INVALID) {
			// Found a matching key, so replace its value
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP(pv_set.vector[index].value);
			index++;
		}
	}

	if (arg->kv_out) {
		arg->kv_out->num_entries = pv_set.num_entries;
		arg->kv_out->vector = pv_set.vector;
		USP_ARG_Init(&pv_set);
	}

	USP_ARG_Destroy(&pv_set);
}

int _get_value_group(vendor_data_t *arg)
{
	struct blob_buf bb = { };
	char *method, *p;
	void *array;
	int i;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "proto", "usp");
	if (strcmp(arg->path, ROOT_PATH) == 0) {
		method = "get";
		blobmsg_add_string(&bb, "path", arg->path);
	} else if (arg->kv->num_entries < MIN_NUM_TO_GROUP) {
		method = "getm_values";
		array = blobmsg_open_array(&bb, "paths");
		for (i = 0; i < arg->kv->num_entries; ++i) {
			p = arg->kv->vector[i].key;
			blobmsg_add_string(&bb, NULL, p);
		}
		blobmsg_close_array(&bb, array);
	} else {
		method = "get";
		blobmsg_add_string(&bb, "path", arg->path);
	}

	uspd_call(g_ubus_ctx, method, &bb, _get_value_group_cb, arg);

	blob_buf_free(&bb);
	return arg->fault;
}

int ubus_enque_cmd(vendor_data_t *arg)
{
	//USP_LOG_Debug("## CMD (%s) path(%s) ##", stringify_cmd(arg->cmd), arg->path);
	if (IS_UBUS_Q_READY) {
		int bytes;

		pthread_mutex_lock(&g_ubus_lock);
		bytes = send(mq_tx_socket, &arg, sizeof(vendor_data_t *), 0);
		if (bytes != sizeof(vendor_data_t *)) {
			pthread_mutex_unlock(&g_ubus_lock);
			USP_LOG_Error("[DMThread] send size miss-match (%d:%zd) to ubus thread", bytes, sizeof(vendor_data_t *));
			return USP_ERR_INTERNAL_ERROR;
		}
		pthread_cond_wait(&g_wait_ubus, &g_ubus_lock);
		pthread_mutex_unlock(&g_ubus_lock);

		return arg->fault;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_TRAN_START:
		_tran_exec(arg, "transaction_start");
		break;
	case CMD_TRAN_COMMIT:
		_tran_exec(arg, "transaction_commit");
		break;
	case CMD_TRAN_ABORT:
		_tran_exec(arg, "transaction_abort");
		break;
	default:
		USP_ERR_SetMessage("%s(%d): uspd thread not avilable", __FUNCTION__, __LINE__);
		arg->fault = USP_ERR_INTERNAL_ERROR;
		break;
	}

	return arg->fault;
}

int uspd_operate_async(dm_req_t *req, kv_vector_t *input_args, int instance)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_OPERATE_ASYNC;
	arg.path = req->path;
	arg.kv = input_args;
	arg.inst = instance;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

int uspd_operate_sync(dm_req_t *req, char *command_key,
		      kv_vector_t *input_args, kv_vector_t *output_args)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_OPERATE_SYNC;
	arg.path = req->path;
	arg.kv = input_args;
	arg.kv_out = output_args;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static void _get_instances_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	vendor_data_t *arg = req->priv;
	struct blob_attr *cur, *params = NULL;
	size_t rem;
	enum {
		P_PARAM,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING }
	};

	if (!msg)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};
		char *path;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (tb[P_PARAM]) {
			path = blobmsg_get_string(tb[P_PARAM]);
			STR_VECTOR_Add(arg->vec, path);
		}
	}
}

static int _get_instances(vendor_data_t *arg)
{
	struct blob_buf bb = {};

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", arg->path);
	blobmsg_add_string(&bb, "proto", "usp");

	uspd_call(g_ubus_ctx, "instances", &bb, _get_instances_cb, arg);

	blob_buf_free(&bb);

	return USP_ERR_OK;
}

static void uspd_register_uniq_param(struct blob_attr *blob_spath, struct blob_attr *uniq_keys)
{
	int num_keys, i;
	size_t rem;
	struct blob_attr *cur;
	char **keys, *spath;

	if (uniq_keys == NULL)
		return;

	spath = blobmsg_get_string(blob_spath);
	num_keys = blobmsg_check_array(uniq_keys, BLOBMSG_TYPE_STRING);
	keys = (char **) USP_MALLOC(num_keys * sizeof(char *));
	if (keys == NULL)
		return;

	i = 0;
	blobmsg_for_each_attr(cur, uniq_keys, rem) {
		if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
			continue;

		keys[i++] = blobmsg_get_string(cur);
	}

	USP_REGISTER_Object_UniqueKey(spath, keys, num_keys);
	USP_FREE(keys);
}

int get_path_from_group_schema(char *group_name, char *schema)
{
	size_t i, len, j;

	j = 0;
	len = strlen(schema);
	group_name[0] = '\0';

	for (i = 0; i < len; i++) {
		if (schema[i] == '{') {
			group_name[j++] = '*';
			i += 2;
		} else {
			group_name[j++] = schema[i];
		}
	}

	// remove the last *
	if (group_name[j - 1] == '*')
		group_name[j - 1] = '\0';

	return 0;
}

bool is_cached_data(kv_vector_t *params)
{
	kv_vector_t *kv;
	int i;
	bool all_found = false;
	kv_pair_t *pair;

	kv = &g_cached_kv;
	if (kv->num_entries == 0)
		return false;

	for (i = 0; i < params->num_entries; i++) {
		pair = &params->vector[i];
		g_cached_index = KV_VECTOR_FindKey(kv, pair->key, g_cached_index);
		if (g_cached_index != INVALID) {
			USP_SAFE_FREE(pair->value);
			pair->value = USP_STRDUP(kv->vector[g_cached_index].value);
			g_cached_index++;
			all_found = true;
		} else {
			all_found = false;
			break;
		}
	}

	return all_found;
}

int group_get(int group_id, kv_vector_t *params)
{
	char group_path[MAX_DM_PATH] = {0};
	vendor_data_t arg;
	bool found = false;

	memset(&arg, 0, sizeof(vendor_data_t));

	found = is_cached_data(params);

	// If all parameters not found in cached data, clean cached data and get it from uspd
	if (found) {
		if (group_id == g_group_vec.num_entries - 1) {
			USP_ARG_Destroy(&g_cached_kv);
		}
		return USP_ERR_OK;
	} else {
		USP_ARG_Destroy(&g_cached_kv);
	}

	get_path_from_group_schema(group_path, g_group_vec.vector[group_id]);
	arg.cmd = CMD_GROUP_GET;
	arg.path = group_path;
	arg.kv = params;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

int uspd_set_value(dm_req_t *req, char *buf)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_SET;
	arg.path = req->path;
	arg.p_value = buf;
	arg.len = strlen(buf);

	ubus_enque_cmd(&arg);

	return arg.fault;
}

int group_set(int group_id, kv_vector_t *params, unsigned *param_types, int *failure_index)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));

	*failure_index = INVALID;
	arg.cmd = CMD_GROUP_SET;
	arg.kv = params;
	arg.p_failure_index = failure_index;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

int group_add(int group_id, char *path, int *instance)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_ADD;
	arg.path = path;
	arg.p_instance = instance;

	ubus_enque_cmd(&arg);

	if (arg.fault != 0)
		return USP_ERR_CREATION_FAILURE;

	return 0;
}

int group_del(int group_id, char *path)
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_DEL;
	arg.path = path;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

int refresh_instances(int group_id, char *path, int *expiry_period)
{
	int i;
	vendor_data_t arg;
	str_vector_t inst_vec;

	memset(&arg, 0, sizeof(vendor_data_t));

	STR_VECTOR_Init(&inst_vec);
	arg.cmd = CMD_INSTANCES;
	arg.path = path;
	arg.vec = &inst_vec;

	// Free cached data
	if (g_cached_kv.num_entries) {
		KV_VECTOR_Destroy(&g_cached_kv);
	}

	ubus_enque_cmd(&arg);

	for (i = 0; i < inst_vec.num_entries; i++) {
		USP_DM_RefreshInstance(inst_vec.vector[i]);
	}

	// cache the object instance numbers for 30 seconds
	// TODO: make it configuration option ??
	*expiry_period = MAX_CACHE_TIME;
	STR_VECTOR_Destroy(&inst_vec);

	return USP_ERR_OK;
}

static void update_instance_refresh_vector(char *path, str_vector_t *refresh_vec)
{
	int i;
	size_t len;
	bool ref_found;

	ref_found = false;
	for (i = 0; i < refresh_vec->num_entries ; i++) {

		len = strlen(refresh_vec->vector[i]);
		if (strncmp(path, refresh_vec->vector[i], len) == 0) {
			ref_found = true;
			break;
		}
	}

	if (ref_found == false)
		STR_VECTOR_Add(refresh_vec, path);
}

static bool create_dm_group(char *spath, int dm_type, bool is_writable, str_vector_t *refresh_vec)
{
	bool ret = false;
	int group_id, index;
	char group_name[MAX_DM_PATH], tmp[MAX_DM_PATH];

	if (CountPathSeparator(spath) < MAX_GROUP_SEP) {
		// USP_LOG_Debug("path(%s) count(%d) is less", spath, MAX_GROUP_SEP);
		return false;
	}

	TEXT_UTILS_SplitPathAtSeparator(spath, group_name, MAX_DM_PATH, 2);
	index = STR_VECTOR_Find(&g_group_vec, group_name);
	// Add group if does not exists
	if (index == INVALID) {
		group_id = g_group_vec.num_entries;
		USP_SNPRINTF(tmp, MAX_DM_PATH, "%s{i}.", group_name);
		// USP_LOG_Debug("\n### ADD GROUP [%d => %s] ###", group_id, group_name);
		if (strcmp(spath, tmp) == 0) {
			USP_REGISTER_GroupedObject(group_id, spath, is_writable);
			update_instance_refresh_vector(spath, refresh_vec);
			ret = true;
		} else {
			USP_REGISTER_GroupedObject(group_id, group_name, 0);
			ret = false;
		}

		USP_REGISTER_GroupVendorHooks(group_id, group_get, group_set, group_add, group_del);
		STR_VECTOR_Add(&g_group_vec, group_name);
	}

	return ret;
}

static void uspd_register_object(char *spath, bool is_writable)
{
	int group, i;

	group = INVALID;

	if (g_group_vec.num_entries == 0) {
		USP_LOG_Error("Create group before object registration");
		return;
	}

	for (i = g_group_vec.num_entries - 1; i >= 0; i--) {
		char *p = g_group_vec.vector[i];
		size_t len = strlen(p);

		if (strncmp(spath, p, len) == 0) {
			group = i;
			break;
		}
	}

	if (group == INVALID) {
		USP_LOG_Error("## REG err (%s)", spath);
	} else {
		// USP_LOG_Debug("## Enroll GROUP [%d => %s]", group, spath);
		USP_REGISTER_GroupedObject(group, spath, is_writable);
	}
}

static int get_dm_type(struct blob_attr *btype)
{
	char *type;

	if (!btype)
		return INVALID;

	type = blobmsg_get_string(btype);
	if (strcmp(type, "xsd:string") == 0)
		return DM_STRING;
	else if (strcmp(type, "xsd:unsignedInt") == 0)
		return DM_UINT;
	else if (strcmp(type, "xsd:int") == 0)
		return DM_INT;
	else if (strcmp(type, "xsd:unsignedLong") == 0)
		return DM_ULONG;
	else if (strcmp(type, "xsd:Long") == 0)
		return DM_ULONG;
	else if (strcmp(type, "xsd:boolean") == 0)
		return DM_BOOL;
	else if (strcmp(type, "xsd:dateTime") == 0)
		return DM_DATETIME;
	else if (strcmp(type, "xsd:object") == 0)
		return DM_VENDOR_MOBJ;
	else
		return DM_STRING;

	return DM_STRING;
}

int uspd_get_value(dm_req_t *req, char *buf, int len)
{
	vendor_data_t arg;
	kv_vector_t kv_param;

	memset(&arg, 0, sizeof(vendor_data_t));
	USP_ARG_Init(&kv_param);

	// This is hack to optimize overall get time
	if (g_cached_kv.num_entries == 0) {
		USP_ARG_Add(&kv_param, req->path, NULL);
		arg.kv_out = &g_cached_kv;
		arg.cmd = CMD_GROUP_GET;
		arg.path = ROOT_PATH;
		arg.kv = &kv_param;
	} else {
		g_cached_index = KV_VECTOR_FindKey(&g_cached_kv, req->path, g_cached_index);
		if (g_cached_index == INVALID) {
			arg.cmd = CMD_GET;
			arg.path = req->path;
			arg.p_value = buf;
			arg.len = len;
		} else {
			USP_STRNCPY(buf, g_cached_kv.vector[g_cached_index].value, len);
			goto exit;
		}
	}

	ubus_enque_cmd(&arg);

	if (kv_param.num_entries && arg.fault == USP_ERR_OK) {
		USP_STRNCPY(buf, kv_param.vector[0].value, len)
	}
exit:
	USP_ARG_Destroy(&kv_param);
	return arg.fault;
}

bool handle_num_entries_reg(char *spath)
{
	char *p;
	int len;
	char param[MAX_DM_PATH];
	int fault = USP_ERR_OK;

	// check if the dm entry is a NumberOfEntries parameter
	len = strlen(spath) - strlen(NUMENTRY);
	if (len < 0)
		return false;

	p = spath + len;
	if (strcmp(p, NUMENTRY) == 0) {
		// prepare dm object with which it should associate
		USP_STRNCPY(param, spath, len + 1);
		strcat(param, ".{i}");
		fault = USP_REGISTER_Param_NumEntries(spath, param);

		if (fault == USP_ERR_OK)
			return true;
	}

	return false;
}

static void uspd_register_leaf(char *spath, bool is_writable, int type)
{
	int group, i;
	char *p;
	size_t len;

	if (handle_num_entries_reg(spath))
		return;

	// most likely the leaf is of last added group, so checking it in
	// reverse search
	group = INVALID;
	for (i = g_group_vec.num_entries - 1; i >= 0; i--) {
		p = g_group_vec.vector[i];
		len = strlen(p);
		if (strncmp(spath, p, len) == 0) {
			group = i;
			break;
		}
	}

	if (group == INVALID) {
		// USP_LOG_Debug("LEAF without group (%s)", spath);
		if (is_writable)
			USP_REGISTER_VendorParam_ReadWrite(spath, uspd_get_value, uspd_set_value, NULL, 0);
		else
			USP_REGISTER_VendorParam_ReadOnly(spath, uspd_get_value, 0);
	} else {
		// USP_LOG_Debug("# LEAF (%d => %s)", group, spath);
		if (is_writable)
			USP_REGISTER_GroupedVendorParam_ReadWrite(group, spath, type);
		else
			USP_REGISTER_GroupedVendorParam_ReadOnly(group, spath, type);
	}

}

bool handled_separately(kv_vector_t *cc_param, char *spath, int dm_type)
{
	char *val;

	if ((strncmp(spath, "Device.BulkData.", 16) == 0) ||
	    (strncmp(spath, "Device.Security.", 16) == 0))
		return true;

	val = KV_VECTOR_Get(cc_param, spath, NULL, 0);
	if (val) {
		USP_REGISTER_Param_Constant(spath, val, dm_type);
		return true;
	}

	return false;
}

static void schema_get_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	vendor_data_t *arg = req->priv;
	kv_vector_t *cc_param = arg->kv;
	str_vector_t ref_param_vec;
	size_t rem, num_uniq, i;
	enum {
		P_PARAM,
		P_WRITEABLE,
		P_TYPE,
		P_UNIQ,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{ "parameter", BLOBMSG_TYPE_STRING },
		{ "writable", BLOBMSG_TYPE_STRING },
		{ "type", BLOBMSG_TYPE_STRING },
		{ "unique_keys", BLOBMSG_TYPE_ARRAY },
	};
	struct uniq_arr {
		struct blob_attr *attr_spath;
		struct blob_attr *uniq_keys;
	} *u_arr;

	if (!msg)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	STR_VECTOR_Init(&ref_param_vec);
	num_uniq = 0;
	u_arr = NULL;
	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX] = {NULL};
		char *dm_spath;
		int dm_type;
		bool is_writable;

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_TYPE] || !tb[P_WRITEABLE])
			continue;

		if (TEXT_UTILS_StringToBool(blobmsg_get_string(tb[P_WRITEABLE]), &is_writable) != USP_ERR_OK) {
			USP_LOG_Info("Decode perm of some param failed");
			continue;
		}

		dm_spath = blobmsg_get_string(tb[P_PARAM]);
		dm_type = get_dm_type(tb[P_TYPE]);

		// store the uniq-keys with schema path, registration should be done after object registration
		if (tb[P_UNIQ]) {
			num_uniq++;
			u_arr = (struct uniq_arr *) USP_REALLOC(u_arr, (num_uniq * sizeof(struct uniq_arr)));
			u_arr[num_uniq - 1].attr_spath = tb[P_PARAM];
			u_arr[num_uniq - 1].uniq_keys = tb[P_UNIQ];
		}

		if (handled_separately(cc_param, dm_spath, dm_type))
			continue;

		if (create_dm_group(dm_spath, dm_type, is_writable, &ref_param_vec))
			continue;

		if (dm_type == DM_VENDOR_MOBJ) {
			update_instance_refresh_vector(dm_spath, &ref_param_vec);
			uspd_register_object(dm_spath, is_writable);
		} else {
			uspd_register_leaf(dm_spath, is_writable, dm_type);
		}
	}

	// register uniq parameters
	for (i = 0; i < num_uniq; ++i) {
		uspd_register_uniq_param(u_arr[i].attr_spath, u_arr[i].uniq_keys);
	}
	USP_SAFE_FREE(u_arr);

	// register refresh instances
	for (i = 0; i < ref_param_vec.num_entries; ++i) {
		// USP_LOG_Debug("REF Inst(%s)", ref_param_vec.vector[i]);
		USP_REGISTER_Object_RefreshInstances(ref_param_vec.vector[i], refresh_instances);
	}
	STR_VECTOR_Destroy(&ref_param_vec);

}

static int uspd_register_args(char *spath, struct blob_attr *in_attr, struct blob_attr *out_attr)
{
	size_t rem;
	int in_nr = 0, out_nr = 0, i, fault = USP_ERR_OK;
	struct blob_attr *cur;
	char **in = NULL, **out = NULL;

	if (in_attr) {
		in_nr = blobmsg_check_array(in_attr, BLOBMSG_TYPE_STRING);
		in = (char **) USP_MALLOC(in_nr * sizeof(char *));
		if (in == NULL)
			return USP_ERR_RESOURCES_EXCEEDED;

		i = 0;
		blobmsg_for_each_attr(cur, in_attr, rem) {
			if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
				continue;

			in[i++] = blobmsg_get_string(cur);
		}
	}

	if (out_attr) {
		out_nr = blobmsg_check_array(out_attr, BLOBMSG_TYPE_STRING);
		out = (char **) USP_MALLOC(out_nr * sizeof(char *));
		if (out == NULL) {
			USP_FREE(in);
			return USP_ERR_RESOURCES_EXCEEDED;
		}

		i = 0;
		blobmsg_for_each_attr(cur, out_attr, rem) {
			if (blobmsg_type(cur) != BLOBMSG_TYPE_STRING)
				continue;

			out[i++] = blobmsg_get_string(cur);
		}
	}

	if (in_nr || out_nr)
		fault = USP_REGISTER_OperationArguments(spath, in, in_nr, out, out_nr);

	USP_FREE(in);
	USP_FREE(out);

	return fault;
}

static int uspd_register_operate_dm(struct blob_attr *param, struct blob_attr *type,
				    struct blob_attr *in, struct blob_attr *out)
{
	char *path;
	int fault = USP_ERR_OK;

	path = blobmsg_get_string(param);
	if (strncmp(blobmsg_get_string(type), "sync", 4) == 0) {
		if ((strncmp(path, "Device.Reboot()", 15) == 0) ||
		    (strncmp(path, "Device.FactoryReset()", 21) == 0))
			return 0;

		fault = USP_REGISTER_SyncOperation(path, uspd_operate_sync);
	} else {
		fault = USP_REGISTER_AsyncOperation(path,
					    uspd_operate_async, NULL);
	}

	if (fault == USP_ERR_OK)
		fault = uspd_register_args(path, in, out);

	return fault;
}

static void operate_schema_get_cb(struct ubus_request *req, int type, struct blob_attr *msg)
{
	struct blob_attr *params;
	struct blob_attr *cur;
	size_t rem;

	enum {
		P_PARAM,
		P_TYPE,
		P_INPUT,
		P_OUTPUT,
		__P_MAX
	};
	const struct blobmsg_policy p[__P_MAX] = {
		{"parameter", BLOBMSG_TYPE_STRING},
		{"type", BLOBMSG_TYPE_STRING},
		{"in", BLOBMSG_TYPE_ARRAY},
		{"out", BLOBMSG_TYPE_ARRAY},
	};

	if (!msg)
		return;

	params = get_parameters(msg);
	if (params == NULL)
		return;

	blobmsg_for_each_attr(cur, params, rem) {
		struct blob_attr *tb[__P_MAX];

		blobmsg_parse(p, __P_MAX, tb, blobmsg_data(cur), blobmsg_len(cur));
		if (!tb[P_PARAM] || !tb[P_TYPE])
			continue;

		uspd_register_operate_dm(tb[P_PARAM], tb[P_TYPE], tb[P_INPUT], tb[P_OUTPUT]);
	}
}

static void read_cmd_callback(struct uloop_fd *u, unsigned int events)
{
	vendor_data_t *arg = NULL;
	int bytes_read;

	pthread_mutex_lock(&g_ubus_lock);
	bytes_read = recv(mq_rx_socket, &arg, sizeof(vendor_data_t *), 0);
	if (bytes_read != sizeof(vendor_data_t *)) {
		USP_LOG_Error("[UBUS] size mismatch %d", bytes_read);
		pthread_cond_signal(&g_wait_ubus);
		pthread_mutex_unlock(&g_ubus_lock);
		return;
	}

	switch (arg->cmd) {
	case CMD_GET:
		_get_value_single(arg);
		break;
	case CMD_GROUP_GET:
		_get_value_group(arg);
		break;
	case CMD_INSTANCES:
		_get_instances(arg);
		break;
	case CMD_ADD:
		_add_object(arg);
		break;
	case CMD_SET:
		_set_value_single(arg);
		break;
	case CMD_GROUP_SET:
		_set_value_group(arg);
		break;
	case CMD_DEL:
		_del_object(arg);
		break;
	case CMD_OPERATE_SYNC:
		_operate_sync_handler(arg);
		break;
	case CMD_OPERATE_ASYNC:
		_operate_async_handler(arg);
		break;
	case CMD_TRAN_START:
		_tran_exec(arg, "transaction_start");
		break;
	case CMD_TRAN_COMMIT:
		_tran_exec(arg, "transaction_commit");
		break;
	case CMD_TRAN_ABORT:
		_tran_exec(arg, "transaction_abort");
		break;
	case CMD_NULL:
		USP_LOG_Info("Command not supported as of now");
		break;
	}

	pthread_cond_signal(&g_wait_ubus);
	pthread_mutex_unlock(&g_ubus_lock);
}

static void *start_ubus_thread(void *arg)
{
	struct uloop_fd cmd_fd;

	uloop_init();
	ubus_add_uloop(g_ubus_ctx);

	memset(&cmd_fd, 0, sizeof(struct uloop_fd));
	cmd_fd.cb = read_cmd_callback;
	cmd_fd.fd = mq_rx_socket;
	uloop_fd_add(&cmd_fd, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	uloop_run();

	uloop_done();

	close(mq_tx_socket);
	close(mq_rx_socket);

	return NULL;
}

static int vendor_exec_init(void)
{
    int err;

    // Exit if unable to initialize the unix domain socket pair used to implement a message queue
    err = socketpair(AF_UNIX, SOCK_DGRAM, 0, uspd_mq_sockets);
    if (err != 0)
    {
        USP_ERR_ERRNO("socketpair", errno);
        return USP_ERR_INTERNAL_ERROR;
    }

    return USP_ERR_OK;
}

static int uspd_register_operate_schema(void)
{

	struct blob_buf bb = {};
	int fault = USP_ERR_OK;

	blob_buf_init(&bb, 0);
	fault = uspd_call(g_ubus_ctx, "list_operate", &bb, operate_schema_get_cb, NULL);
	blob_buf_free(&bb);

	return fault;
}

static int uspd_register_schema(vendor_data_t *arg)
{
	struct blob_buf bb = {};
	int fault = USP_ERR_OK;

	// Invoke Ubus to get data from uspd
	blob_buf_init(&bb, 0);
	fault = uspd_call(g_ubus_ctx, "dump_schema", &bb, schema_get_cb, arg);
	blob_buf_free(&bb);

	if (fault == USP_ERR_OK)
		fault = uspd_register_operate_schema();

	return fault;
}

int factory_reset(void)
{
	int res = USP_ERR_OK;
	struct blob_buf bb;

	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", ROOT_PATH);
	blobmsg_add_string(&bb, "action", "FactoryReset()");

	res = uspd_call_inline_nocb("operate", &bb);
	blob_buf_free(&bb);

	return res;
}

int reboot(void)
{
	int res = USP_ERR_OK;
	struct blob_buf bb;


	memset(&bb, 0, sizeof(struct blob_buf));
	blob_buf_init(&bb, 0);

	blobmsg_add_string(&bb, "path", ROOT_PATH);
	blobmsg_add_string(&bb, "action", "Reboot()");

	res = uspd_call_inline_nocb("operate", &bb);
	blob_buf_free(&bb);

	return res;
}

static int uspd_tran_start()
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_START;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int uspd_tran_commit()
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_COMMIT;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int uspd_tran_abort()
{
	vendor_data_t arg;

	memset(&arg, 0, sizeof(vendor_data_t));
	arg.cmd = CMD_TRAN_ABORT;

	ubus_enque_cmd(&arg);

	return arg.fault;
}

static int vendor_hook_init(void)
{
	vendor_hook_cb_t callbacks;

	memset(&callbacks, 0, sizeof(callbacks));
	callbacks.reboot_cb = reboot;
	callbacks.factory_reset_cb = factory_reset;
	callbacks.start_trans_cb = uspd_tran_start;
	callbacks.commit_trans_cb = uspd_tran_commit;
	callbacks.abort_trans_cb = uspd_tran_abort;

	return USP_REGISTER_CoreVendorHooks(&callbacks);
}

void uspd_cache_mandatory_data(vendor_data_t *arg)
{
	kv_vector_t *kv = arg->kv;

	USP_ARG_Add(kv, "Device.DeviceInfo.SoftwareVersion", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.ProductClass", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.Manufacturer", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.ModelName", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.HardwareVersion", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.ManufacturerOUI", "");
	USP_ARG_Add(kv, "Device.DeviceInfo.SerialNumber", "");

	arg->cmd = CMD_GROUP_GET;
	arg->path = "Device.DeviceInfo.";

	ubus_enque_cmd(arg);
}

// Public methods
int vendor_uspd_init(void)
{
	int fault = USP_ERR_OK;
	kv_vector_t cached_kv;
	vendor_data_t arg;

	g_ubus_ctx = ubus_connect(NULL);
	if (!g_ubus_ctx) {
		USP_LOG_Error("[%s:%d] ubus_connect failed", __func__, __LINE__);
		return USP_ERR_INTERNAL_ERROR;
	}

	// register mandatory parameters
	memset(&arg, 0, sizeof(vendor_data_t));
	KV_VECTOR_Init(&cached_kv);
	arg.cmd = CMD_NULL;
	arg.kv = &cached_kv;

	uspd_cache_mandatory_data(&arg);

	// Default group for rest of the parameters
	KV_VECTOR_Init(&g_cached_kv);
	STR_VECTOR_Init(&g_group_vec);

	fault = uspd_register_schema(&arg);
	if (fault == USP_ERR_OK)
		fault = vendor_hook_init();
	else
		vendor_uspd_stop();

	KV_VECTOR_Destroy(&cached_kv);

	return fault;
}

int vendor_uspd_stop(void)
{
	int i;

	// Need to free manually as memory allocated in Init
	for (i = 0; i < g_group_vec.num_entries; i++) {
		free(g_group_vec.vector[i]);
	}
	if (g_group_vec.vector)
		free(g_group_vec.vector);
	STR_VECTOR_Init(&g_group_vec);

	KV_VECTOR_Destroy(&g_cached_kv);
	ubus_free(g_ubus_ctx);
	g_ubus_ctx = NULL;
	return USP_ERR_OK;
}

int vendor_uspd_start(void)
{
	int fault = USP_ERR_OK;

	fault = vendor_exec_init();
	if (fault == USP_ERR_OK)
		fault = OS_UTILS_CreateThread(start_ubus_thread, NULL);

	return fault;
}
